# File txmbox.tcl
#
# Provides an advanced replacement for tk_messageBox.
# Procs that are declared PUBLIC in the TOC are user callable.
# All other procs are supporting.
# Tabs (for editing this file) are set to 4.
#
# This package will eat memory unless the user calls the
# clean routine after the popup information is no longer needed.
#
#
#<< doctcl -t 4 -v 1.0.1
#
# See ChangeLog.html for changes
#       2006.09.02  Robert Joy  New
#
# Table of Contents
#
# detab							PUBLIC
# alignText						PUBLIC
# cleanMsgBox					PUBLIC
# pushGrabStack					PUBLIC
# popGrabStack					PUBLIC
# waitForClose					PUBLIC
# posGeom
# destroyOnPidDeath
# getMsgBoxGeometry
# destroyMsgBox					PUBLIC
# buttonPress					
# updateMsgBox
# changeMsgBoxButtonLabel		PUBLIC
# addToMsgBox					PUBLIC
# configButton					PUBLIC
# messageBox					PUBLIC
# button3Popup					PUBLIC
# menuCmd						PUBLIC
# isWaiting
# placeWindow					PUBLIC
# centerWindow					PUBLIC
# scrollSet
#
#...............................................................................
if { [info exists ::txmbox::version] } { return }

# We now require the Bwidget package
#
package require BWidget

namespace eval txmbox {

	variable	version 1.0.6
	variable	TxEnv
	variable	TxMessageBox
	array set	TxMessageBox { LastName "" }

	namespace export messageBox
	namespace export destroyMsgBox
	namespace export configButton
	namespace export addToMsgBox
	namespace export changeMsgBoxButtonLabel
	namespace export cleanMsgBox
	
	namespace export TxMessageBox
	
	package provide txmbox $version
	package provide Txmbox $version
	
}


# See if Tile was loaded by the user
#
if { [ regexp {Tile} [info loaded] ] } {
	set ::txmbox::TxEnv(tile) 1
} else {
	set ::txmbox::TxEnv(tile) 0
}

# Set platform dependent params
#
set ::txmbox::TxEnv(isUnix) 0
set ::txmbox::TxEnv(isMac)	0
set ::txmbox::TxEnv(isWin)	0

switch $tcl_platform(platform) {
	"unix"      { set ::txmbox::TxEnv(isUnix)	1}
	"windows"   { set ::txmbox::TxEnv(isWin)	1; package require twapi }
	"macintosh" { set ::txmbox::TxEnv(isMac)	1}
}

set ::txmbox::TxEnv(ImagePath)	[ file join [ file dirname [ info lib ] ] txmbox$::txmbox::version images ]
foreach ::txmbox::icon [list question error warning info] {
	
	set ::txmbox::TxEnv($::txmbox::icon)	[ image create photo -format GIF \
											-file [ file join $::txmbox::TxEnv(ImagePath) ${::txmbox::icon}.gif ] ]
	
}
	

#------------------------------------------------------------------------- detab
#
# convert \t characters to blanks in given body of text
#
# offset is the number of chars already written to the first line
#
#	the tab argument gives the tab field width, default is 8.
#
proc ::txmbox::detab { text {tab 8} {offset 0} } {

	set rtn {}

	if {$tab < 2} {
		
		subst -all "\t" $text " " text
		return $text
		
	}

	if { [ regexp "^\[ \n]*$" $text ] } { return "" }; 		# blank

	set sep ""
	
	foreach line [ split $text "\n" ] {
		
		# pos is the number of chars already written to the current line
		#
		set pos    $offset
		
		set offset 0
		
		set buff  ""
		
		#puts "\nLINE: pos $pos, offset $offset, Split line: <[split $line \t]>"
		
		foreach part [ split $line \t ] {
			
			if {"$part" == "" && "$buff" == ""} {
				
				#
				# This is the empty space before a tab 
				# at the start of a new line.
				# We will do another tab when we get the next part
				#
				if {$pos == 0} {set part " "}
				
				#puts "   EMPTY part at start: pos $pos; part '$part'>"
				
			} elseif { $pos > 0 } {	
				
				set width [ expr {$tab - ($pos % $tab)} ] 
				
				set part [ format "%${width}s%s" "" $part ]
				
				#puts "   pos $pos, width $width, part '$part'"
			}
			
			append buff $part
			
			incr pos [ string length $part ]
		}
		
		append rtn $sep $buff
		
		set sep "\n"
		
	}

	return $rtn

}


#-------------------------------------------------------- alignText
#
# This only works correctly with fixed width fonts. The
# returned string will have mixed results with variable
# width fonts.
#
# Aligns text to a given page width so it can later be displayed
# as a nicely formatted block of text. The following rules apply:
#
# 1. Text that extends beyond the page width will wrap at the
#    last space that occurs just before the page edge.
#
# 2. Any line that is not butting up against the left edge
#    will be considered indented. The indent amount will be detected
#    and any wrapping text will include this additional indent. The
#    routine will attempt to handle things like indented bullets
#    by detecting the number of chars from the start of the bullet
#    to the start of the actual text for that bullet. All
#    but the first line shall have an additional indent equal to
#    the number of chars detected. If no bullet is detected the
#    wrapped lines will align with the first line. The following
#    initial chars (when followed by some white space) will constitute
#    a bullet:
#
#	 *, -, +, N., N)
#
#    where N is any number. Examples:
#
#		* This is a bullet
#		- So is this
#		1.      And this
#		2)	this one too
#
#   The user is expected to handle the left edge of the text. What
#   is meant here is, if the user wants numbered bullets that are
#   aligned around the "." then the user will need to adjust the
#   indent accordingly, ie;
#
#	 1. First line is here. You can see that the
#	    wrapped part is in line with the first
#	    char of the text.
#	14. This might be the fourteenth line. Notice
#	    how both it's bullet and the text aligns
#	    with the previous line.
#
#   The first line above has an additional space before it this allows
#   the periods in the numbers to sit in the same column.
#
# INPUT:	text		- The text to be formatted
#			width		- The width of the page
#			* lmargin	- The left edge of the text
#			* nonewline	- Remove the last newline at the end of the formatted text.
#			* noindents	- Treat indents as regular text.
#
# OUTPUT:	returns a string containing the formatted text
#
proc ::txmbox::alignText { text width { lmargin 0 } { nonewline 0 } { noindents 0 } } {

	set new 		""
	set textlines	[split $text \n]
	set textlcnt	[llength $textlines]

	for { set i 0 } { $i < $textlcnt } { incr i } {
		
		set line [ detab [lindex $textlines $i] ]
		
		# Deal with an indent
		set indent 0
		
		if { [regexp {^( +).*} $line dummy spaces] &&
			 ! $noindents					} {
			
			set indent [string length $spaces]
			
		} else {
			
			set indent 0
			
		}
		
		# Now remove the indent from the line
		set line [string trimleft $line]
		
		set done		0
		set start		0
		set foundsome	1
		set nextindent	0
		
		
		while { ! $done } {
			
			set section [string range $line $start end]
			set sectlen [string length $section]
			
			if { [expr $sectlen + $indent + $nextindent + $lmargin] > $width } {
				
				set end [string last " " [string range $line 0 [ expr $start + $width - $nextindent - $indent - $lmargin - 1]]]
				if { $end < 0 ||
					 $end < $start } {
					
				  set end [expr $start + $width - $nextindent - $indent - $lmargin - 1]
					
				}
				
			} else {
				
				set end  [expr [string length $line] - 1]
				set done 1
				
			}
			
			append new [format "%-*s%s\n" [expr $indent + $nextindent + $lmargin] "" [string range $line $start $end]]
			
			# All wrapped indented lines will be indented an additional 2 spaces
			
			if { $indent > 0 &&
				 $nextindent == 0	} {
				
				if { [regexp {^((\*|-|\+) +).*$} $line dummy head] } {
					
					set nextindent [string length $head]
					
				} elseif { [regexp {^([0-9]+(.|\)) +).*$} $line dummy head] } {
					
					set nextindent [string length $head]
					
				}
				
			}
			
			set start [expr $end + 1]
			
		}

	}

	if { $nonewline } { set new [string range $new 0 [expr [string length $new] - 2]] }
	return $new

}


#-------------------------------------------------------- pidExists
#
# INPUT:	pid - the process id number
#
# RETURNS:	1 - if process is still running
#		    0 - if process is no longer running
#
proc ::txmbox::pidExists { pid { name 0 } } {

	variable TxEnv
	
    set all ""

    # Assume pid does not exist if we have an error
    #
	if { $TxEnv(isWin) } {
		
		if { $name } {
			
			if { [ twapi::get_process_ids -name $pid ] == "" } { return 0 }
			
		} else {
			
			if { ! [ twapi::process_exists $pid ] } { return 0 }
			
		}
		
	} else {
		
		if [ catch { set all [ exec ps -p $pid ] } result ] {
			
		  return 0
			
		}
		
		if { [ regexp {.*defunct.*} $all ] } {
			
		  return 0
			
		}
		
	}

    return 1

}


#------------------------------------------------------------ cleanMsgBox
#
# Recovers variable space used when creating and manipulating
# txmbox popup windows.
#
# INPUT:	* w - the name of the widget to grab
#
proc ::txmbox::cleanMsgBox { {w ""} } {

	variable TxMessageBox
	
	if { $w == "" } {
		
		# Make sure we only get the ones that are dialog specific
		#
		set AllVars [ array names txmbox::TxMessageBox *.txmbox* ]
		
	} else {
		
		set AllVars [ array names txmbox::TxMessageBox *$w ]
		
	}
	
	foreach var $AllVars {
		
		regexp {\.txmbox.*$} $var w
		
		if { [ winfo exists $w ] } { continue }
		catch { unset txmbox::TxMessageBox($var) }
		
	}
	
}


#------------------------------------------------------------ PushGrabStack
#
# Push the current grab onto the grab stack
#
# INPUT:	w		- the name of the widget to grab
#			global	- grab globally
#
proc ::txmbox::pushGrabStack { w { global 0 } } {

	variable TxEnv

	catch {
		
		lappend TxEnv(GRABSTACK) [grab current]
		
		if { $global } {
			
			grab -global $w
			
		} else {
			
			grab $w
			
		}
		
	}

}


#------------------------------------------------------------ PopGrabStack
#
# Pop the last grab off the grab stack
#
proc ::txmbox::popGrabStack {} {

  variable TxEnv

	if {[info exists TxEnv(GRABSTACK)]} {
		
		if { [lrange $TxEnv(GRABSTACK) end end] != "" &&
			 [lrange $TxEnv(GRABSTACK) end end] != "{}"} {
			
			catch { grab set [lrange $TxEnv(GRABSTACK) end end] }
			set TxEnv(GRABSTACK) [lreplace $TxEnv(GRABSTACK) end end]
			
		} elseif {[lrange $TxEnv(GRABSTACK) end end] == "" ||
				  [lrange $TxEnv(GRABSTACK) end end] == "{}"} {
			
			set TxEnv(GRABSTACK) [lreplace $TxEnv(GRABSTACK) end end]
			
		}
		
	}

}


#------------------------------------------------------------ waitForClose
#
# Grab the focus and wait for the window to close
#
# INPUT:	w - window that we are waiting for
#
proc ::txmbox::waitForClose { w { focus "" } { pid "" } { force 1 } { parent "" }} {

	variable TxEnv

	if { $focus != "" } {
		
		if { $force } {
			
			focus -force $focus
			
		} else {
			
			focus $focus
			
		}
		
	} else {
		
		if { $force } {
			
			focus -force  $w
			
		} else {
			
			focus $w
			
		}
		
	}

	update idletasks
	txmbox::pushGrabStack $w
  
	# Find the window that last had the grab
	#
	if { $parent == "" } {
		
		if { [info exists TxEnv(GRABSTACK)] &&
			 $TxEnv(GRABSTACK) != "{}" &&
			 $TxEnv(GRABSTACK) != "" } {
			
			set win [lindex $TxEnv(GRABSTACK) end]
			
		} else {
			
			set win .
			
		}
		
	} else {
		
		set win $parent
		
	}
  
	# make this window a transient of the one before it
	# This works assuming the window before was either . or
	# had pushed onto the Grab Stack
	#
	wm transient $w $win
  
	#catch { raise $w }
  
	if { $pid != "" } {
		
		txmbox::destroyOnPidDeath $w $pid
		
	}
  
	# the catch is just in case the window gets destroyed before we get here
  
	catch { tkwait window $w }
  
	txmbox::popGrabStack

}


#------------------------------------------------------------ posGeom
#
# Finds the relative geometry for the window based on + reference
#
proc ::txmbox::posGeom { w } {

	regexp -- {[0-9]+x(\+|-|)[0-9]+(\+[0-9]+|-[0-9]+)(\+[0-9]+|-[0-9]+)} [wm geometry $w] dummy1 dummy2 X Y
	
	return ${X}${Y}

}


#------------------------------------------------------------ destroyOnPidDeath
#
# Destroys the widget w when the pid is no longer present in
# process listing
#
proc ::txmbox::destroyOnPidDeath { w pid } {

	variable TxMessageBox
	
	if { ! [ txmbox::pidExists $pid ] } {
		
		txmbox::destroyMsgBox $w
		
	} else {
		
		after 2000 [list txmbox::destroyOnPidDeath $w $pid]
		
	}

}


#------------------------------------------------------------ getMsgBoxGeometry
#
# Gets the geometry of the txmbox window w when called
#
proc ::txmbox::getMsgBoxGeometry { w } {

	variable TxMessageBox
	
	catch {
		
		if { [ wm state [winfo parent $w] ] == "normal" } {
			
			#puts "GEOM MODE = $TxMessageBox(GeometryMode.$w)"
			
			switch -exact $TxMessageBox(GeometryMode.$w) {
				
				xy		{ set tmp [txmbox::posGeom $w] }
				full	{
					
					set text	${w}.mf.fr2.fr; #.mess
					set xy		[txmbox::posGeom $w]
					set width	[winfo width $text]
					set height	[winfo height $text]
					set grid	[wm grid $w]
					set gw		[expr ( $width / [lindex $grid 2] )]
					set gh		[expr ( $height / [lindex $grid 3] )]
					#puts "W = $width, H = $height, GRID = $grid"
					set tmp 	"${gw}x${gh}$xy"
					
				}
				none		{ set tmp "" }
				
			}
			
			if { [info exists TxMessageBox(GeometrySet.$w)] &&
				 $tmp != $TxMessageBox(GeometrySet.$w)			} {
				
				set TxMessageBox(Geometry.$w)	$tmp
				
			}
			
		}
		
	}

}


#------------------------------------------------------------ destroyMsgBox
#
# Destroys the widget w when called
#
proc ::txmbox::destroyMsgBox { w } {

	variable TxMessageBox

	catch {
		
		txmbox::getMsgBoxGeometry $w
		
		destroy $w
		
	}
	
	if { $TxMessageBox(Callback.$w) != "" } {
		
		$TxMessageBox(Callback.$w) ""
		
	}

}


proc ::txmbox::buttonPress { button } {

	variable TxMessageBox
	
	set parts [split $button "."]
	
	set w .[lindex $parts 1]
	
	if { ! [ winfo exists $w ] } { return }
	
	if { $TxMessageBox(Confirm.$w) } {
		
		set response	[ txmbox::messageBox	-title		"Confirm" \
												-beep		true \
												-message	"Are you sure you want to do this?" \
												-icon		question \
												-font		$TxMessageBox(ConfirmFont.$w) \
												-type		yesno \
												-wait		1 \
												-parent		$w \
						]
		
		if { $response == "no" } { return }
		
	}
	
	txmbox::getMsgBoxGeometry $w
	
	# The delay here is needed to compensate for a bug in the BWidget Button
	after 10 [list destroy $w]
	
	set TxMessageBox(Button.$w)			[ $button cget -text ]
	set TxMessageBox(ButtonPressed.$w)	1
	
	if { $TxMessageBox(Callback.$w) != "" } {
		
		$TxMessageBox(Callback.$w) $TxMessageBox(Button.$w)
		
	}

}


#------------------------------------------------------------ updateMsgBox
#
# Update timed windows seconds
#
proc ::txmbox::updateMsgBox { w remaining title } {

	variable TxMessageBox
	
	if { [winfo exists $w] } {
		
		wm title $w "$title  [format {(expires in %6d secs)} $remaining]"
	  
		incr remaining -1
	  
		if {$remaining < 0} {
			
			txmbox::destroyMsgBox $w
			
		} else {
			
			after 1000 [list txmbox::updateMsgBox $w $remaining $title]
			
		}
		
	}

}


#------------------------------------------------------------ changeMsgBoxButtonLabel
#
# Changes the label on the indicated button.
#
proc ::txmbox::changeMsgBoxButtonLabel { widget name new } {

	catch {
		
		set namep	[cleanButtonPath $name]
		set newp	[cleanButtonPath $new]
		
		if { [ winfo exists ${widget}.of ] } {
			
			set w ${widget}.of.bf.bf.$namep
			
		} else {
			
			set w ${widget}.bf.bf.$namep
			
		}
		
		$w configure -text $new -width [expr [string length $new] + 1]
		
	}
}


#------------------------------------------------------------ addToMsgBox
#
# Provides a mechanism of adding information to the textbox widget inside
# the window that was returned when messageBox was called with
# the -wait = false. messageBox will handle the additional data
# such that if the data extends beyond the currently available
# space then a vertical scrollbar will appear and the window will
# scroll automatically as new data is entered. If the scrollbar is
# moved by the user so it's no longer sitting at the bottom
# of the window then automatic scrolling will stop for as long as
# the scrollbar is not at the bottom.
#
proc ::txmbox::addToMsgBox { widget msg {fg ""} {bg ""} } {

	set dotag 0
	
	if { [winfo exists ${widget}.of] } {
		
		set w ${widget}.of.mf.fr2.fr.mess
		
	} else {
		
		set w ${widget}.mf.fr2.fr.mess
		
	}
	
	if { $fg != "" || $bg != "" } {
		
		set clk [clock seconds][clock clicks]
		if { [ set idx [string first "-" $clk] ] >= 0 } {
			
		  set clk [string range $clk 0 [expr $idx - 1]]q[string range $clk [expr $idx + 1] end]
			
		}
	  
		# Make tag name
		#
		set tagname "tag$clk"
	  
		set dotag 1
		
	}
	
	# Do this just in case the window goes away while we are trying to update it
	#
	if [ catch {
		
		if { $fg != "" && $bg != "" } {
			
			$w tag configure $tagname -background $bg -foreground $fg
			
		} elseif { $fg != "" } {
			
			$w tag configure $tagname -foreground $fg
			
		} elseif { $bg != "" } {
			
			$w tag configure $tagname -background $bg
			
		}
		
		# find out where the scrollbar is
		set where [$w yview]
		
		$w config -state normal
		
		# Get text end index
		#
		set start [$w index "end -1 char"]
		
		$w insert end $msg
		
		if { $dotag } { $w tag add $tagname $start "end -1 char" }
		
		update idletasks
		
		# if the scrollbar was at the bottom then move the view with each new line
		if { [lindex $where 1] == 1 } {
			
			$w yview moveto 1.0
			
		}
		
		$w config -state disabled
		
	} result ] {
		
		if { [regexp {invalid command name} $result] } {
			
			return 0
			
		} else {
			
			bgerror $result
			return 0
			
		}
		
	}
	
	return 1

}


#--------------------------------------------------------------- configButton
#
# Configures the various aspects of a button so we have
# a consistent look and feel around the GUI.
#
# This routine must be called after the creation of a button
#
proc ::txmbox::configButton { w { hl 1 } } {

	variable TxMessageBox

	if { [info exists TxMessageBox(Button.relief)]		} { $w configure	-relief				$TxMessageBox(Button.relief) }
	if { [info exists TxMessageBox(Button.fg)]			} { $w configure	-fg					$TxMessageBox(Button.fg) }
	if { [info exists TxMessageBox(Button.bg)]			} { $w configure	-bg					$TxMessageBox(Button.bg) }
	if { [info exists TxMessageBox(Button.activefg)]	} { $w configure	-activeforeground	$TxMessageBox(Button.activefg) }
	if { [info exists TxMessageBox(Button.activebg)]	} { $w configure	-activebackground	$TxMessageBox(Button.activebg) }
	if { [info exists TxMessageBox(Button.hlwidth)] &&
		 $hl											} { $w configure	-highlightthickness	$TxMessageBox(Button.hlwidth) }

}


#--------------------------------------------------------------- cleanButtonPath
#
# Configures the various aspects of a button so we have
# a consistent look and feel around the GUI.
#
# This routine must be called after the creation of a button
#
proc ::txmbox::cleanButtonPath { button } {

	variable TxMessageBox

	set lb [string tolower $button]
	regsub -all {[ \.\$\{\}\[\]:]} $lb {a} lb
	
	return $lb
	
}


#------------------------------------------------------------ messageBox
#
# Pops up an info dialog window that can be set to timeout in N
# seconds. It returns it's path name when -wait is false (default)
# OR the button label of the pressed button or an empty string (timeout) when -wait
# is true. Text can be dynamically added to the message content via the routine
# addToMsgBox. It adheres to all the tk_messageBox options.
# This dialog box routine was created to improve existing
# tk_messageBox capabilities and add many addtional ablilities.
#
# The dialog will add a vertical scroll bar if the text height exceeds
# 25 lines. By default the width of the window is computed based on the
# widest lines in the message up to a maximum of 100 chars. Lines longer
# than 100 chars are wrapped.
#
# Unlike tk_messageBox, messageBox uses a text widget to hold the
# message. As a result, the contents of the message can be copied and
# pasted to a file. messageBox disables the widget so the user can't
# alter the text.
#
#
# The following txmbox variables are available:
#
#	$TxMessageBox(Button.$widget)		- The name of the button pressed.
#						  				  Empty string if no button was pressed.
#						  				  Useful when the -wait option is false.
#
#	$TxMessageBox(ButtonPressed.$widget)- Flag indicating the button was pressed or not.
#
#	$TxMessageBox(Geometry.$widget)		- If the window was repostioned by the user
#						  				  contains the xy geometry when the window closed
#						  				  else contains a null string.
#						  				  Useful if you want to save the user's preferred
#						  				  window placement.
#
#	TxMessageBox(Font)					- Allows the font to be set for all messageBoxes
#						  				  which have no -font option.
#
#	TxMessageBox(FontSize)				- Allows the font size to be set for all messageBoxes
#						  				  which have no -font option.
#
# The following global variable are honored for buttons only and assumng you're not using Tile:
#
#	TxMessageBox(Button.fg)			- foreground color if not otherwise specified
#	TxMessageBox(Button.bg)			- background color if not otherwise specified
#	TxMessageBox(Button.relief)		- relief; default is normal button default
#	TxMessageBox(Button.hlwidth)	- highlight width default is 0
#	TxMessageBox(Button.activefg)	- foreground color when mouse if over button
#	TxMessageBox(Button.activebg)	- background color when mouse if over button
#
#
# OPTIONS:
#		message		- Message to be displayed in dialog.
#
#		title		- The text in the titlebar (Default is ATTENTION).
#
#		before		- A proc to call before creating the dialog.
#
#		timeout		- How long to wait before automatically closing window.
#					  0 waits forever. Default is 0.
#
#		showtime	- For timed windows, adds time remaining to titlebar
#					  of timed dialogs.
#
#		type		- Valid values are:
#
#					    ok, okcancel, cancel, yesno, yesnocancel, abort,
#					    abortretrycancel, retrycancel, none, custom
#
#					  If type is custom then a list of button names
#					  must be passed in using the typebuttons option. The button
#					  name will appear on the buttons just as you have specified
#					  them. Currently only 1 row of buttons will be made.
#
#					  Default is ok.
#
#		typebuttons	- A list of custom button names. The value returned will be
#					  in lowercase.
#
#		beep		- Gives an audible alert when window popsup.
#
#		image		- A gif to place on the left side of the window.
#					  Default is no image.
#
#		icon		- Name of the icon to use. If none specified
#					  none will be displayed. -image will override
#					  this option if it occurs later in the arg list.
#
#		again		- A global variable to hold the state of the "show again"
#					  checkbutton.
#
#		default		- The default button (green highlight if not using tile). Default is first button
#					  in type.
#
#		wait		- Wait for the user to respond to the dialog then
#					  return the button that was pressed.
#					  Valid args are: 0, 1, true, false, yes, no
#					  Default is true.
#
#		width		- Force the width to this number of characters.
#					  If width is not specified then the window width will be
#					  set to the smaller of 100 columns or computed number
#					  of columns to display based on all other options provided.
#
#		height		- Force the height to this number of lines.
#					  If height is not specified then the window height will be
#					  set to the smaller of 25 lines or computed number of lines to display
#					  based on all other options provided. In this case, if the number of resulting
#					  lines is greater than 25 then a vertical scrollbar will be added.
#
#		watch		- Watch a pid and for completion then close dialog
#					  and return an empty string. If user presses a button
#					  then return button name.
#
#		callback	- A process to call with the name of the button that was pressed.
#
#		xy			- Specify the XY coordinates for the window. Valid
#					  forms are:
#
#						+X+Y, +X-Y, -X+Y, -X-Y
#
#		geom		- This option may be used as an alternative to specifying the
#					  xy, height, width options. The geometry mode will be set to
#					  full. The geometry variable will then report the full geometry
#					  for the window instead of just the xy position. If no geom is
#					  specified then the window will be placed at the screen center.
#
#		font		- A fully qualified font specification to use on all
#					  textual aspects of the dialog. This should be a fixed
#					  width font, ie;
#
#						{helvetic 16 normal italic}
#
#		force		- If true then forces the focus to the default button
#					  or the window if type "none" is specified. Default
#					  is to force focus.
#
#		grab		- Grabs the focus so the parent is unable to process
#					  any keys or buttons. Keeps the event loop running.
#					  Default is to not grab.
#
#		resize		- If resize is true then the window can be resized by
#					  the user otherwise not. Default is true.
#
#		notitle		- If this option is true then no window dressing will be
#					  applied to the popup. This is useful for "Please Wait" screens.
#					  Default is false.
#
#		buttons		- Passes the name of a global variable where messageBox
#					  will create a list of eval type expressions that will allow the
#					  message box buttons to be executed remotely. The list 
#					  should be accessed in the following fashion:
#
#							eval [lindex $<var name> <button number>]
#
#		name		- Passes the name of a global variable where messageBox
#					  will save the name of the messageBox toplevel window.
#
#		wicon		- Provide the absolute path to the xbm file that should be used for
#					  iconifed window. This is mainly for systems where the icon will be seen
#					  in a taskbar or icon window. Default is no icon.
#
#		format		- If format is true then the message will use the rules for the
#					  alignText routine and they are:
#
#						Aligns text to the width of the window so it can later be displayed
#						as a nicely formatted block of text. The following rules apply:
#
#						1. Text that extends beyond the page width will wrap at the
#						   last space that occurs just before the page edge.
#
#						2. Any line that is not butting up against the left edge
#						   will be considered indented. The indent amount will be detected
#						   and any wrapping text will include this additional indent. The
#						   routine will attempt to handle things like indented bullets
#						   by detecting the number of chars from the start of the bullet
#						   to the start of the actual text for that bullet. All
#						   but the first line shall have an additional indent equal to
#						   the number of chars detected. If no bullet is detected the
#						   wrapped lines will align with the first line. The following
#						   initial chars (when followed by some white space) will constitute
#						   a bullet:
#
#						      *, -, +, N., N)
#
#						   where N is any number. Examples:
#
#					              * This is a bullet
#					              - So is this
#					              1.      And this
#					              2)      this one too
#
#						  The user is expected to handle the left edge of the text. What
#						  is meant here is, if the user wants numbered bullets that are
#						  aligned around the "." then the user will need to adjust the
#						  indent accordingly, ie;
#
#						       1. First line is here. You can see that the
#						          wrapped part is in line with the first
#						          char of the text.
#						      14. This might be the fourteenth line. Notice
#						          how both it's bullet and the text aligns
#						          with the previous line.
#
#						  The first line above has an additional space before it in order
#						  for the periods in the numbers to sit in the same column.
#
#					  The width that will be used for formatting will be the value passed
#					  in by the -width option OR, if that's not provided, the length of the
#					  longest line present in the message OR 100 which ever is less.
#
#					  This option should only be used in cases where the
#					  text is predefined by the programmer as resizing the window can cause
#					  the text to become ugly and difficult for the user to read as a result
#					  the minsize for the window will be fixed to the calculated maximum
#					  width and height for the text window. Default is no formatting.
#
#		parent		- Specifies the transient leader for this new dialog. By default the leader
#					  is ".".
#
#		confirm		- Present a confirmation window when the user clicks any button on the messageBox.
#					  Default is to NOT confirm.
#
#		nolast		- Do not set the LastButton variable when this window is created. This mainly
#					  to prevent sub-messageBox windows from tainting this txmbox::TxMessageBox(LastName)
#
# RETURNS:	window name or button pressed
#
proc ::txmbox::messageBox { args } {

	variable	TxMessageBox
	variable	TxEnv
	global		errorInfo tk_strictMotif
	
	update idletasks
	
	set windowingsystem [ tk windowingsystem ]
	set fontsAvailable	[ string tolower [ font families ] ]
	
	set clk [clock seconds][clock clicks]
	if { [ set idx [string first "-" $clk] ] >= 0 } {
		
		set clk [string range $clk 0 [expr $idx - 1]]q[string range $clk [expr $idx + 1] end]
		
	}
	set w .txmbox$clk
	
	set TxMessageBox(Button.$w)			""
	set TxMessageBox(ButtonPressed.$w)	0
	set TxMessageBox(Callback.$w)		""
	set TxMessageBox(Geometry.$w)		""
	set TxMessageBox(GeometryMode.$w)	none
	
	set message			""
	set title			"ATTENTION!!"
	set before			""
	set timeout			0
	set type			ok
	set typebuttons		""
	set default			ok
	set image			""
	set icon			""
	set useicon			1
	set again			""
	set wait			1
	set width			0
	set height			0
	set resize			1
	set notitle			0
	set showtime		0
	set format			0
	set confirm			0
	set watch			""
	set xy				""
	set geom			""
	set buttons			""
	set name			""
	set wicon			""
	set beep			0
	# Fix font spec for windows if needed
	if { $TxEnv(isWin) } {
		
		if { [info exists TxMessageBox(FontSize)] } {
			
			set fontsize	$TxMessageBox(FontSize)
			
		} else {
			
			set fontsize	10
			
		}
		
		if { [ info exists TxMessageBox(Font) ] &&
			 [ lsearch -exact $fontsAvailable [ string tolower $TxMessageBox(Font) ] ] >= 0 } {
			
			set fontname	$TxMessageBox(Font)
			
		} else {
			
			set fontname	"courier new"
			
		}
		
	} else {
		
		if { [info exists TxMessageBox(FontSize)] } {
			
			set fontsize	$TxMessageBox(FontSize)
			
		} else {
			
			set fontsize	12
			
		}
		
		if { [ info exists TxMessageBox(Font) ]  &&
			 [ lsearch -exact $fontsAvailable [ string tolower $TxMessageBox(Font) ] ] >= 0 } {
			
			set fontname	$TxMessageBox(Font)
			
		} else {
			
			set fontname	"fixed"
			
		}
		
	}
	set font					[list $fontname $fontsize normal roman]
	set focus					$w
	set force					1
	set grab					0
	set parent					.
	set state					flag
	
	
	set options					[list	message default icon parent title type \
										image beep again timeout wait width showtime \
										watch callback xy font force grab height \
										resize notitle name buttons geom format wicon \
										typebuttons confirm before \
								]
	set types					[list	ok okcancel cancel yesno yesnocancel abort \
										abortretrycancel retrycancel none custom \
								]
	set icons					[list	error info question warning]
	set erroroption				"Invalid option passed to txmbox::messageBox:"
	set errorvalue				"Invalid value passed to txmbox::messageBox:"
	set errormissing			"Missing option value txmbox::messageBox:"
	set errorcombo				"Invalid combination of values passed to txmbox::messageBox:"
	set typeary(ok)				[list Ok]
	set typeary(okcancel)		[list Ok Cancel]
	set typeary(cancel)			[list Cancel]
	set typeary(yesno)			[list Yes No]
	set typeary(yesnocancel)	[list Yes No Cancel]
	set typeary(abort)			[list Abort]
	set typeary(abortretrycancel)	[list Abort Retry Cancel]
	set typeary(retrycancel)	[list Retry Cancel]
	
	foreach arg $args {
		
		switch -exact -- $state {
			
			flag	{
				
				set state value
				set option [string range $arg 1 end]
				
				if { [lsearch -exact $options $option] < 0 } {
					
					bgerror "$erroroption $arg"
					return 0
					
				}
				
			}
			
			value	{
				
				set state flag
				set $option $arg
				
				# do value check and set related vars
				switch -exact -- $option {
					
					message	{ # no additional required }
					
					default	{ # this must be checked after all values set }
					
					icon	{
						
						set useicon 1
						if { [lsearch -exact $icons [set $option]] < 0 } {
							
							bgerror "$errorvalue -$option [set $option]\n  Must be one of \"$icons\""
							return 0
							
						}
						
					}
					
					title	{ # no additional }
					
					before	{
						
						set procName ::[ lindex [set $option] 0 ]
						if { [ info procs $procName ] == "" } {
							
							bgerror "$errorvalue -$option [set $option]\n  Proc \"$procName\" does not exist"
							return 0
							
						}
						
					}
					
					type	{
						
						if { [lsearch -exact $types [set $option]] < 0 } {
							
							bgerror "$errorvalue -$option [set $option]\n  Must be one of \"$types\""
							return 0
							
						}
						
					}
					
					typebuttons	{ # no additional }
					
					image	{ set useicon 0 }
					
					again	{ # no additional }
					
					watch	{
						
						if { [regexp {$[0-9]+^} [set $option]] } {
							
							bgerror "$errorvalue -$option [set $option]\n  Must be a positive integer"
							return 0
							
						}
						
						set wait 1
						
					}
					
					height -
					width -
					timeout	{
						
						if { [regexp {$[0-9]+^} [set $option]] } {
							
							bgerror "$errorvalue -$option [set $option]\n  Must be a positive integer"
							return 0
							
						}
						
					}
					
					callback	{
						
						set TxMessageBox(Callback.$w) $arg
						
					}
					
					beep		-
					confirm		-
					format		-
					notitle		-
					resize		-
					grab		-
					force		-
					showtime	-
					wait	{
						
						set arg [string tolower $arg]
						
						switch -exact -- $arg {
							
							0 -
							no -
							false	{ set $option 0 }
							
							1 -
							yes -
							true	{ set $option 1 }
							
						}
						
					}
					
					font	{
						
						if { [llength $font] < 4 } {
							
							bgerror "$errorvalue -$option [set $option]\n  Invalid number of font specifiers"
							return 0
							
						}
						
					}
					
					xy		{
						
						if { ! [regexp {(^$|^\+[0-9]+\+[0-9]+$|^\+[0-9]+-[0-9]+$|^-[0-9]+\+[0-9]+$|^-[0-9]+-[0-9]+$)} [set $option]] } {
							
							bgerror "$errorvalue -$option [set $option]\n  Must be one of the following:\n\n\t+X+Y\n\t+X-Y\n\t-X+Y\n\t-X-Y"
							return 0
							
						}
						
						set TxMessageBox(GeometryMode.$w) xy
						
					}
					
					geom	{
						
						if { ! [regexp {(^$|^[0-9]+x[0-9]+(\+[0-9]+\+[0-9]+|\+[0-9]+-[0-9]+|-[0-9]+\+[0-9]+|-[0-9]+-[0-9]+)$)} [set $option]] } {
							
							bgerror "$errorvalue -$option [set $option]\n  Must be one of the following:\n\n\tWxH+X+Y\n\tWxH+X-Y\n\tWxH-X+Y\n\tWxH-X-Y"
							return 0
							
						}
						
						set TxMessageBox(GeometryMode.$w) full
						
					}
					
					buttons	{
						
						upvar #0 [set $option] BUTTONS
						
					}
					
					name	{
						
						upvar #0 [set $option] NAME
						
					}
					
					parent	{
						
						if { ! [winfo exists [set $option]] } {
							
							bgerror "$errorvalue -$option [set $option]\n  Invalid window name. Window does not exist."
							return 0
							
						}
						
					}
					
					wicon	{
						
						if { ! [file exists [set $option]] } {
							
							bgerror "$errorvalue -$option [set $option]\n  File does not exist."
							return 0
							
						}
						
					}
					
					
				}
				
			}
			
		}
		
	}
	
	if { $state == "value" } {
		
		# Then the user gave a flag but no value
		bgerror "$errormissing -$option"
		return 0
		
	}

	# Make sure we don't have a locking combination of options
	#
	if { $wait &&
		(
			( $type == "none" ) ||
			( $type == "custom" && $typebuttons == "" )
		)													} {
		
		bgerror "$errorcombo\n  A locking combination of button options exists."
		return 0
		
	}
	
	# If we're using a customer button set then make sure the caller specified the buttons
	#
	if { $type != "none" } {
		
		if { $type == "custom" } {
			
			if { $typebuttons == "" } {
				
				bgerror "$errorvalue -typebuttons\n  If -type is custom then button names must also be specified using the -typebuttons option."
				return 0
				
			} else {
				
				set typeary(custom) $typebuttons
				
			}
			
		}
		
	}
	
	# Find out if we need to support Tile
	#
	if { [ regexp {Tile} [info loaded] ] } {
		set tile 1
		set TxEnv(tile) 1
	} else {
		set tile 0
		set TxEnv(tile) 0
	}

	set max			15
	set hmax		1
	set scrollbar	0
	set lines		[split $message \n]
	set lcnt		[llength $lines]

	# Try to guess at the correct width for the text widget
	#
	if { $width == 0 } {
		
		# No width was provided by the user so find out
		# how long the longest line is. If it's less than
		# 100 then use it, if not, then set it to 100.
		#
		foreach line $lines {
			
			set l [string length [detab $line]]
			
			if { $l > $max } { set max $l }
			
		}
		
		if { $max < 100 } {
			
			set newWidth $max
			
		} else {
			
			set newWidth 100
			
		}
		
	} else {
		
		if { $again != "" } {
			
			set newWidth [ expr ( $width < 30 )?30:$width ]
			
		} else {
			
			set newWidth $width
			
		}
		
		if { $type != "none" } {
			
			set wb 0
			foreach button $typeary($type) {
				
				if { $type == "custom" } {
					
					set bw [string length $button]
					
				} else {
					
					set bw 6
					
				}
				
				incr wb $bw
				incr wb 2
				
			}
			
			set newWidth [ expr ($newWidth < $wb)?$wb:$newWidth ]
			
		}
		
	}

	if { $format } {
		
		set newtext [alignText $message $newWidth 0 1]
		set message	$newtext
		
	} else {
		
		set newtext [alignText $message $newWidth 0 1 1]
		
	}
  
	set linecnt [llength [split $newtext \n]]
  
	if { $height != 0 } {
		
		set scrollbar	0
		set hmax		$height
		set max			$newWidth
		
	} elseif { $linecnt > 25 } {
		
		set scrollbar	1
		set hmax		25
		set max			$newWidth
		
	} else {
		
		set scrollbar	0
		set hmax		$linecnt
		set max			$newWidth
		
	}


	# See if we need to call a before proc
	#
	if { $before != "" } { catch { eval $before } }
	

	# Ok, make the damn thing already
	#
	toplevel			$w -bd 0 -highlightthickness 0
  
	if { $notitle 		} { wm overrideredirect $w 1 }
	wm title			$w $title
	wm transient		$w $parent
	wm protocol			$w WM_DELETE_WINDOW {# do nothing}
	wm resizable		$w $resize $resize
	if { $format		} { wm minsize $w $max $hmax }
	if { $wicon != ""	} { wm iconbitmap $w @$wicon }
  
	set realw $w

	# Due to differences in the way window managers function
	# we must do this for non-MSWindows platforms
	#
	if { ! $TxEnv(isWin) } {
		
		wm withdraw			$realw
		update
		
	}
	
	if { $name != "" } { set NAME $w }
	set TxMessageBox(LastName) $w
	

  
	# If no title then create an additional frame so we can
	# display a window with beveled edges. For Tile this may
	# result in a window with no bevels. Depends on the theme.
	#
	if { $notitle } {
		
		if { $tile } {
			ttk::frame $w.of -borderwidth 4
		} else {
			frame $w.of -bd 4 -relief raised
		}
		pack $w.of -expand true -fill both -ipadx 0
		
		set w $w.of
		
	}
  
	if { $tile } {
		ttk::frame $w.mf -borderwidth 0
	} else {
		frame $w.mf -bd 0
	}
	pack  $w.mf -side top -expand true -fill both -ipadx 0

	#.......................................................................... Add icon or image
	#
	if { $useicon } {
		
		# Mac aqua icons
		#
		if {$windowingsystem == "aqua"} {
			
			switch -- $icon {
				
				"error"     {set icon "stop"}
				"warning"   {set icon "caution"}
				"info"      {set icon "note"}
				
			}
			
		}
		
		if { $icon != "" } {
			
			if { $tile } {
				ttk::frame $w.mf.fr1 -borderwidth 0
			} else {
				frame $w.mf.fr1 -bd 0
			}
			pack  $w.mf.fr1 -side left -ipadx 0
			
			# For color limited windowing systems
			#
			if { $windowingsystem == "aqua" || ( [winfo depth $w] < 5 ) || $tk_strictMotif } {
				
				if { $tile } {
					ttk::label $w.mf.fr1.image -bitmap $icon -text 1 -bd 0 -padx 0
				} else {
					label $w.mf.fr1.image -bitmap $icon -text 1 -bd 0 -padx 0
				}
				
			} else {
				
				if { $tile } {
					ttk::label $w.mf.fr1.image -image $TxEnv($icon) -text 1
				} else {
					label $w.mf.fr1.image -image $TxEnv($icon) -text 1 -bd 0 -padx 0
				}
				
			}
			pack  $w.mf.fr1.image -side top -padx 4 -pady 4 -ipadx 0
			
		}
		
	} else {
		
		if { $image != "" } {
			
			if { $tile } {
				ttk::frame $w.mf.fr1 -borderwidth 0
			} else {
				frame $w.mf.fr1 -bd 0
			}
			pack  $w.mf.fr1 -side left  -ipadx 0
			
			if { $tile } {
				ttk::label $w.mf.fr1.image -image $image -text 1 -borderwidth 0 -padding {0 0 0 0}
			} else {
				label $w.mf.fr1.image -image $image -text 1 -bd 0 -padx 0
			}
			pack  $w.mf.fr1.image -side top -padx 4 -pady 4 -ipadx 0
			
		}
		
	}

	if { $tile } {
		ttk::frame $w.mf.fr2 -borderwidth 0
	} else {
		frame $w.mf.fr2 -bd 0
	}
	pack	$w.mf.fr2 -side left -fill both -expand true -ipadx 0
  
	if { $tile } {
		ttk::frame $w.mf.fr2.fr -borderwidth 0
	} else {
		frame $w.mf.fr2.fr -bd 0
	}
	pack	$w.mf.fr2.fr -side top -fill both -expand true -padx 6 -pady 6 -ipadx 0
	
	#.......................................................................... Add message widget
	#
	set text [ text $w.mf.fr2.fr.mess	-width $max \
										-height $hmax \
										-wrap word \
										-bd 1 \
										-pady 6 \
										-padx 8 \
										-relief solid \
										-highlightthickness 0 \
										-setgrid true \
										-font $font \
			 ]
	bind $text <Enter>			[list focus $text]
	bind $text <ButtonPress-3>	[list txmbox::button3Popup $text]

	if { $tile } {
		ttk::scrollbar $w.mf.fr2.fr.scroll -command [list $text yview]
	} else {
		scrollbar $w.mf.fr2.fr.scroll -command [list $text yview]
	}
	pack $text -side left -fill both -expand true -padx 2 -pady 2 -ipadx 0

	$text insert end $message
	$text config -state disabled

	# Bind the up, down, pageup, and pagedn keys to the text widget's scroll function
	#
	bind $w <Up>	[list $text yview scroll -1 units]
	bind $w <Down>	[list $text yview scroll  1 units]
	bind $w <Prior>	[list $text yview scroll -1 pages]
	bind $w <Next>	[list $text yview scroll  1 pages]
	bind $w <Home>	[list $text yview moveto 0.0]
	bind $w <End>	[list $text yview moveto 1.0]
	#bind $w <KeyPress>	{puts stdout {%%K=%K  %%A=%A}}


	#.......................................................................... Add the again checkbox?
	#
	if { $again != "" } {
		
		if { $tile } {
			ttk::checkbutton $w.mf.fr2.again	-variable $again \
												-text "Don't show this dialog again"
			set fagain	{"courier new" 12}
			
		} else {
			checkbutton $w.mf.fr2.again	-variable $again \
										-text "Don't show this dialog again" \
										-font $font
			set fagain	[$w.mf.fr2.again cget -font]
			
		}
	  
		pack $w.mf.fr2.again -side top -pady 0 -padx 4
		
	}


	#.......................................................................... Add the buttons
	#
	if { $type != "none" } {
		
		if { $tile } {
			set mbf [ttk::frame $w.bf -borderwidth 0]
		} else {
			set mbf [frame $w.bf -bd 0]
		}
		pack $mbf -side top -pady 0 -ipadx 0 -fill x
		
		
		if { $tile } {
			set bf [ttk::frame $mbf.bf -borderwidth 0]
		} else {
			set bf [frame $mbf.bf -bd 0]
		}
		pack $bf -side top -pady 0 -ipadx 0
		
		
		# .......................................................................... Make the buttons
		#
		if { $buttons != "" } {
			
			set BUTTONS ""
			
		}
		
		foreach button $typeary($type) {
			
			if { $type == "custom" } {
				
			  set bwidth [string length $button]
				
			} else {
				
			  set bwidth 6
				
			}
			
			set lb [cleanButtonPath $button]
			
			if { $tile } {
				
				ttk::button $bf.$lb -text $button \
									-width $bwidth \
									-command [list txmbox::buttonPress $bf.$lb] \
									-state disabled
				
			} else {
				
				Button $bf.$lb	-text $button \
								-width $bwidth \
								-command [list txmbox::buttonPress $bf.$lb] \
								-font $font \
								-highlightthickness 3 \
								-highlightcolor green \
								-state disabled
				configButton $bf.$lb 0
				
			}
			# if { $notitle } { $bf.$lb config -relief flat }
			
			pack   $bf.$lb -side left -padx 6 -pady 10
			bind   $bf.$lb <Return> [list txmbox::buttonPress $bf.$lb]
			
			if { $buttons != "" } {
				
				lappend BUTTONS "txmbox::buttonPress $bf.$lb"
				
			}
			
		}
		
		# Set the default button
		#
		if { [lsearch -exact [string tolower $typeary($type)] [string tolower $default]] >= 0 } {
			
			set focus $bf.[string tolower $default]
			
		} else {
			
			set focus $bf.[string tolower [lindex $typeary($type) 0]]
			
		}
		
	}

  
  
	# Since we add the scrollbars after the initial display
	# we do not need to factor their dimensions into the mix.
  
	set TxMessageBox(Leader.$realw)	[wm transient $realw]
  
	set TxMessageBox(Waiting.$realw)	0
  
	bind $TxMessageBox(Leader.$realw) <Unmap> { txmbox::isWaiting }
  
	if { $TxEnv(isWin) } { wm withdraw $realw }
	update
	
	set fwide [winfo width $realw]
	set fhigh [winfo height $realw]
	
	switch -exact $TxMessageBox(GeometryMode.$realw) {
		
		full	{ placeWindow $realw $geom }
		xy		{ placeWindow $realw $xy }
		none	{ centerWindow $realw $fwide $fhigh }
		
	}

	bind $realw <Configure>		[list txmbox::getMsgBoxGeometry $realw]

	if { $beep } {
		
		if { $TxEnv(isWin) } {
			
			after 300 { twapi::play_sound SystemExclamation -alias }
			
		} else {
			
			bell
			
		}
		
	}
	
	if { $type != "none" } {
		
		# Enable the buttons
		foreach button $typeary($type) {
			
			set lb [cleanButtonPath $button]
			
			if { $tile } {
				$bf.$lb state !disabled
			} else {
				$bf.$lb configure -state normal
			}
			
		}
		
		set TxMessageBox(Confirm.$realw)		$confirm
		set TxMessageBox(ConfirmFont.$realw)	$font
		
	}
  
	set TxMessageBox(GeometrySet.$realw)		[txmbox::posGeom $realw]
  
	if { $scrollbar } { pack $w.mf.fr2.fr.scroll -side right -fill y }
	$text configure -yscrollcommand [list ::txmbox::scrollSet $w.mf.fr2.fr.scroll [list pack configure $w.mf.fr2.fr.scroll -before $w.mf.fr2.fr.mess -fill y -side right]]

	if { $timeout > 0 } {
		
		if { $showtime } {
			
			after 100 [list txmbox::updateMsgBox $realw $timeout $title]
			
		} else {
			
			after [expr $timeout * 1000] [list txmbox::destroyMsgBox $realw]
			
		}
		
	}
  

	if { $wait } {
		
		txmbox::waitForClose $realw $focus $watch $force $parent
		update idletasks
		return $TxMessageBox(Button.$realw)
		
	} else {
		
		if { $force } {
			
			focus -force $focus
			
		} else {
			
			focus $focus
			
		}
		
		if { $grab } {
			
			txmbox::pushGrabStack $realw
			bind $realw <Destroy> [list txmbox::popGrabStack]
			
		}
		
		update idletasks
		return $realw
		
	}

}


#--------------------------------------------------------------- button3Popup
#
# Create the button 3 popup menu based on selection.
#
proc ::txmbox::button3Popup { w } {

	global globs widget
	
	catch { destroy .txmboxb3pop }
	
	set range [$w tag nextrange sel 1.0]
	
	set m [menu .txmboxb3pop -tearoff 0]
	
	if { $range == "" } {
		
		# Do this if nothing has been selected
		$m add command -label "Select All"		-command [list txmbox::menuCmd $w selectAll]
		
	} else {
		
		$m add command -label "Copy"			-command [list txmbox::menuCmd $w copy]
		
	}

    if {0} {
	$m add separator
    $m add command -label "Print Screen..."		-command [list txmbox::menuCmd $w printScreen]
    $m add command -label "Print Selection..."	-command [list txmbox::menuCmd $w printSelection]
	}
	
    tk_popup $m [winfo pointerx $w] [winfo pointery $w]

}


#------------------------------------------------------------------------------- menuCmd
#
# Create the main gui for the tool
#
proc ::txmbox::menuCmd { w cmd } {
	
    global widget globs TxEnv env Mail icon Info
	
	switch -exact -- $cmd {
		
		selectAll	{ $w tag add sel 1.0 end }
		
		copy		{
			
			if [ catch { set val [selection get] } result ] {
				
			    # Ignore
				
			} else {
				
			    clipboard clear
			    clipboard append -- $val
				
			}
			
		}

	}

}


#---------------------------------------------------------------- isWaiting
#
# There is a pending tkwait for this window
#
proc ::txmbox::isWaiting {} {

	variable TxMessageBox
	
	# get the indexes for all waiting windows ( usually only one )
	
	set idxes [array names TxMessageBox Waiting*]
	
	
	# loop through the windows to find if one is waiting
	
	foreach idx $idxes {
		
		# save the window's name
		set w [string range [file extension $idx] 1 end]
	  
		if { $TxMessageBox($idx) == 1 } {
			
			regsub -- Waiting $idx Leader lidx
			
			while { ! [winfo ismapped $w] } {
				
				if { [wm state $TxMessageBox($lidx)] == "iconic" } {
					
				  wm deiconify $TxMessageBox($lidx)
					
				  update idletasks
					
				}
				
			}
			
			bind $TxMessageBox($lidx) <Unmap> {}
			
			wm iconify $TxMessageBox($lidx)
			
			unset TxMessageBox($lidx)
			
			break
			
		}
		
	}

}


proc ::txmbox::placeWindow { w geom } {

	variable TxMessageBox
  
	catch { wm geometry $w $geom }
  
	if { ( [wm transient $w] == "" && [ wm state $w ] == "normal" ) ||
		 ( [wm transient $w] != "" && ( [ wm state [wm transient $w] ] == "normal" || [ wm state [wm transient $w] ] == "zoomed" ) ) } {
		
		catch { wm deiconify $w }
		set   TxMessageBox(Waiting.$w) 1
		
		catch { tkwait visibility $w }
		set   TxMessageBox(Waiting.$w) 0
		
		update idletasks
		
		catch { wm geometry $w $geom }
		
		update
		
	}

}


proc ::txmbox::centerWindow { w { width 100 } { height 100 } } {

	variable TxMessageBox
  
	set sw [winfo screenwidth  .]
	set sh [winfo screenheight .]
  
	catch { wm geometry $w +[expr $sw/2 - $width/2]+[expr $sh/2 - $height/2] }
  
	if { ( [wm transient $w] == "" && [ wm state $w ] == "normal" ) ||
		 ( [wm transient $w] != "" && ( [ wm state [wm transient $w] ] == "normal" || [ wm state [wm transient $w] ] == "zoomed" ) ) } {
		
		catch { wm deiconify $w }
		set   TxMessageBox(Waiting.$w) 1
		
		# TO USE THE ABOVE 2 LINES YOU MUST COMMENT OUT THE NEXT LINE
		
		catch { tkwait visibility $w }
		set   TxMessageBox(Waiting.$w) 0
		
		update idletasks
		
		catch { wm geometry $w +[expr $sw/2 - [winfo width $w]/2]+[expr $sh/2 - [winfo height $w]/2] }
		
		update
		
	}

}


#---------------------------------------------------------------- scrollSet
#
# Manages optional scrollbars.
#
proc ::txmbox::scrollSet { scrollbar geoCmd offset size } {

	if {$offset != 0.0 || $size != 1.0} {
		
		eval $geoCmd                                ;# Make sure it is visible
		$scrollbar set $offset $size
		
	} else {
		
		set manager [lindex $geoCmd 0]
		$manager forget $scrollbar                  ;# hide it
		
	}

}
